# PWBus - RabbitmqClient Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sun Nov 17 20:10:24 -03 2019

import pika
from json import dumps, loads, load
from time import sleep
import traceback

from pwbus.commons.logging import *
from pwbus.clients.client import Client
from pwbus.connectors.rabbitmq import RabbitmqConnector

# RabbitmqClient
#
#


class RabbitmqClient(RabbitmqConnector, Client):

    def __init__(self, host=None, port=None):
        super().__init__(host=host, port=port)
        self.response = None
        self.correlation_id = None

    def setCorrelationId(self, correlation_id):
        self.correlation_id = correlation_id

    def onResponse(self, ch, method, props, body):
        if self.correlation_id == props.correlation_id:
            self.response = body

    # RabbitmqClient.get
    #
    def get(self, resource_name, correlation_id, wait=0.05, retries=20):
        self.setCorrelationId(correlation_id)
        log_debug(
            f'🔎 RabbitmqClient.get - Retrieving response from [{resource_name}] with correlation_id [{correlation_id}] wait={wait}ms retries={retries}')
        try:
            with self.getConnectionPool().item() as connection:
                channel = connection.channel()
                channel.basic_consume(
                    queue=resource_name,
                    on_message_callback=self.onResponse,
                    auto_ack=True
                )

                retry = 0
                while self.response is None and retry < retries:
                    connection.process_data_events()
                    sleep(wait)
                    retry += 1

                channel.stop_consuming()

                if self.response is None:
                    log_debug(
                        f'RabbitmqClient.get - Response not found - correlation_id [{correlation_id}]')
                else:
                    self.response = loads(self.response)
                    log_debug(
                        f'RabbitmqClient.get - Response found - correlation_id [{correlation_id}] retry [{retry+1}]')

                return self.clear_header(self.response)

        except:
            log_error(
                traceback, f'RabbitmqClient.get - Failed to get message by correlation_id')

        return None



# PWBus - EngineManager Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Tue Nov 19 20:07:35 -03 2019

import sys
import importlib
import traceback

from pwbus.commons.logging import *
from pwbus.engines.engine_registry_loader import RegistryLoader
from pwbus.engines.engine_registry import Registry
from pwbus.engines.engine_monitor import EngineMonitor

# EngineManager
#
#


class EngineManager():

    def __init__(self, registry_file):
        # Load registry entries
        self.registry_loader = RegistryLoader(registry_file)
        self.registry_loader.load()

    # EngineManager.action
    #
    def action(self, command, engine, channel):
        running_threads = []

        if not engine:
            engines = self.registry_loader.getRunnableEngines()
        else:
            engines = [engine]

        if engines == []:
            log_debug(
                'EngineManager.action - All engines are disabled, check engine.enabled property in pwbus_registry.json')
            return

        # Start pwbus_monitor
        EngineMonitor().start()

        # Running engines threads
        try:
            for engine_entry in engines:
                module_name = f'pwbus.engines.{engine_entry}.engine'
                module = importlib.import_module(module_name)
                class_ = getattr(module, f'{engine_entry.capitalize()}Engine')

                # channel is None run all channels else only the channel specified
                if not channel:
                    runnable_channels = self.registry_loader.getRunnableChannels(
                        engine=engine_entry)
                else:
                    runnable_channels = [channel]

                registry = Registry()

                log_debug(
                    f'\n\n=====({engine_entry} engines startup)===============================\n')
                for channel_entry in runnable_channels:
                    channel_registry = registry.getChannel(channel_entry)
                    if channel_registry["engine.enabled"]:
                        if not channel_registry["engine.start_threads"]:
                            channel_registry["engine.start_threads"] = 2

                        engine_instance = class_()
                        log_debug(
                            f'EngineManager.action - Starting threads for engine=[{engine_entry}] class=[{engine_entry.capitalize()}Engine] channel=[{channel_entry}]')
                        running_threads += engine_instance.serve(
                            channel_registry)

            # for started_thread in running_threads:
            #   started_thread.join()

        except KeyboardInterrupt:
            log_debug(
                'EngineManager.action - Please wait, stopping engines threads')

            for started_thread in running_threads:
                started_thread.do_run = False
                log_debug(
                    f'EngineManager.action - Stopping thread [{started_thread.name}]')

            log_fatal("EngineManager.action - Interrupted by user")

        except:
            log_error(
                traceback, 'EngineManager.action - Error starting engines threads')
            log_fatal("EngineManager.action - stopped")



# PWBus - MongoConnector Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sat Dec 14 11:32:03 -03 2019

import pymongo
import traceback

from pwbus.commons.logging import *
from pwbus.connectors._connector import _Connector

# MongoConnectorIn
#
#


class MongoConnector(_Connector):

    def __init__(self, host='mongodb', port=21017):
        super().__init__('mongo', host, port)
        self.setConnectionPool(
            self.create_connection,
            max_size=10,
            max_usage=1000,
            idle=300,
            ttl=600
        )

    # MongoConnector.create_connection
    #
    def create_connection(self):
        try:
            return pymongo.MongoClient(f'mongodb://root:example@{self.getHost()}/')
        except:
            log_error(
                traceback, 'MongoConnector - Error on try connect to mongodb server')


# MongoConnectorIn
#
#
class MongoConnectorIn(MongoConnector):

    def __init__(self, host, port):
        super().__init__(host=host, port=port)

# MongoConnectorOut
#
#


class MongoConnectorOut(MongoConnector):

    def __init__(self, host, port):
        super().__init__(host=host, port=port)



# PWBus - HttpConnectorIn and HttpConnectorOut Classes
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sat Nov 16 16:36:36 -03 2019

from pwbus.connectors._connector import ConnectorIn, ConnectorOut

# HttpConnectorIn
#
#


class HttpConnectorIn(ConnectorIn):

    def __init__(self, host=None, port=None):
        super().__init__(connector_type='http-in', host=host, port=port)


# HttpConnectorOut
#
#
class HttpConnectorOut(ConnectorOut):

    def __init__(self, host=None, port=None):
        super().__init__(connector_type='http-out', host=host, port=port)



# PWBus - _Transform Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sun Nov 17 14:07:58 -03 2019

import json
import traceback

from pwbus.commons.logging import *


class _Transform():

    # _Transform.parse_in - receive <any> return <dict>
    #
    def parse_in(self, payload):
        pass

    # _Transform.parse_out - receive <dict> return <any>
    #
    def parse_out(self, payload):
        pass

    # _Transform.jsonToString
    #
    def stringToJson(self, data):
        try:
            if isinstance(data, str):
                return json.loads(data)
            if isinstance(data, dict):
                return data
            else:
                log_debug(
                    f'_Transform.stringToJson - Invalid type [{type(data)}]', data)
                return None

        except json.decoder.JSONDecodeError:
            log_debug("_Transform.jsonToString - Invalid JSON")
            log_message_dump('_Transform.jsonToString', payload_in)

    # _Transform.jsonToString
    #
    def jsonToString(self, data):
        if isinstance(data, dict):
            return json.dumps(data)
        if isinstance(data, str):
            return data
        else:
            log_debug(
                f'_Transform.jsonToString - Invalid type [{type(data)}]', data)
            return None



#!/usr/bin/env python
from setuptools import setup, find_namespace_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name="pwbus",
    version="0.1.27",
    author="Fabio Szostak",
    author_email="fszostak@gmail.com",
    description="An integration bus capable of performing message transformation and managing task execution. Message exchange between plataforms, like HTTP, Redis, RabbitMQ, Apache Kafta, AWS SQS, MongoDB.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/fszostak/pwbus",
    package_dir={"": "src"},
    packages=find_namespace_packages(where="src"),
    entry_points={
        'console_scripts': ['pwbus=pwbus.cli.command_line:main'],
    },
    install_requires=requirements,
    setup_requires=['flake8'],
    classifiers=[
        'Development Status :: 4 - Beta',
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7'
)



# PWBus - Config Class
#:  Load json file
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sat Nov 16 16:36:36 -03 2019

from json import load


class Config():

    def __init__(self, filename):
        with open(filename) as json_file:
            self.config_data = load(json_file)

    # Config.get
    #
    def get(self, name):
        if name in self.config_data:
            return self.config_data[name]
        return None



# PWBus - KafkaClient Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Wed Dec 11 07:13:13 -03 2019

from __future__ import division
import math
from itertools import islice
from pykafka.common import OffsetType
from pykafka.exceptions import SocketDisconnectedError, LeaderNotAvailable
from json import dumps, loads, load
from time import sleep
import traceback

from pwbus.commons.logging import *
from pwbus.clients.client import Client
from pwbus.connectors.kafka import KafkaConnector

# KafkaClient
#
#


class KafkaClient(KafkaConnector, Client):

    def __init__(self, host=None, port=None):
        super().__init__()

    # KafkaClient.get
    #
    def get(self, resource_name, correlation_id, wait=0.05, retries=20):

        log_debug(
            f'🔎 KafkaClient.get - Retrieving response from [{resource_name}] with correlation_id [{correlation_id}] wait={wait}ms retries={retries}')
        try:
            with self.getConnectionPool().item() as connection:
                topic = connection.topics[resource_name]

                LAST_N_MESSAGES = 50

                consumer = topic.get_simple_consumer(
                    consumer_group=resource_name,
                    queued_max_messages=LAST_N_MESSAGES,
                    consumer_timeout_ms=int(wait*retries*1000),
                    auto_commit_enable=True
                )

                offsets = [(p, op.next_offset - LAST_N_MESSAGES - 1)
                           for p, op in consumer._partitions.items()]
                offsets = [(p, o) if o != -1 else (p, -2) for p, o in offsets]
                consumer.reset_offsets(offsets)

                try:
                    consumer.consume()
                except (SocketDisconnectedError) as e:
                    log_error(traceback, "Kafka.get - consumer failed")
                    return None

                for message in consumer:
                    response = loads(''.join(message.value.decode('utf-8')))
                    if 'Pwbus-Correlation-Id' in response and \
                            correlation_id == response['Pwbus-Correlation-Id']:

                        partition = topic.partitions[0]
                        consumer.reset_offsets([(partition, message.offset)])
                        consumer.commit_offsets()

                        log_debug(
                            f'KafkaClient.get - Response found - correlation_id [{correlation_id}]]')
                        return self.clear_header(response)

                log_debug(
                    f'KafkaClient.get - Response not found - correlation_id [{correlation_id}]')

        except:
            log_error(
                traceback, f'KafkaClient.get - Failed to get message by correlation_id')

        return None



# PWBus - DynamicTask Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Mon Nov 18 09:30:57 -03 2019

import importlib
import os.path
import traceback

from pwbus.commons.logging import *

# DynamicTask
#
#


class DynamicTask():

    def __init__(self, task_id, payload, isDebugEnabled=False):
        try:
            task_name = task_id.split('.')
            module_name = f'pwbus_tasks.{task_name[0]}.{task_name[1]}'
            class_name = task_name[1].capitalize()
        except:
            log_debug(
                f'🟥 DynamicTask - Invalid task_id [{task_id}] specify "module.class" (lowercase)')
            payload['errorOnTask'] = f'Invalid task_id [{task_id}]'
            return

        try:
            module = importlib.import_module(module_name)
            class_ = getattr(module, class_name)
            self.instance = class_(payload)

            if isDebugEnabled:
                log_debug(
                    f'DynamicTask - Module [{module_name}] Class [{class_name}] instance created')

        except:
            formatted_lines = traceback.format_exc().splitlines()
            log_debug(
                f'⚠️ DynamicTask - WARNING!!! Cannot instatiate class -  task_id [{task_id}] - Message: {formatted_lines[-2]} | {formatted_lines[-1]}')
            payload['errorOnTask'] = formatted_lines[-1]

            traceback.print_exc()
            raise

    # DynamicTask.getInstance
    #
    def getInstance(self):
        return self.instance

    # DynamicTask.isLoaded
    #
    def isLoaded(self):
        return True if self.instance else False
